/*
********************************************************************
(c) MSU Video Group 2003-2006, http://compression.ru/video/
This source code is property of MSU Graphics and Media Lab

This code may be distributed under LGPL
(see http://www.gnu.org/licenses/lgpl.html for more details).

E-mail: video-measure@compression.ru
Author: Alexey Noskov
********************************************************************
*/  

/*
* bi_psnr.cpp: implementation of VMT plugin for measurement of brigyhness-independent PSNR.
*/

#include "stdafx.h"

#include "pluginadapter.h"
#include "bi_psnr.h"

template < typename T > T sqr ( const T & a ) { return a * a; }

/*
* DllMain
*/

BOOL APIENTRY DllMain ( HANDLE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved )
{
    return TRUE;
}

extern "C" __declspec(dllexport) void CreateMetric ( IMetricPlugin**metric )
{
	*metric = new CPluginAdapter ( new CBIPSNRPlugin );
}

extern "C" __declspec(dllexport) void ReleaseMetric ( IMetricPlugin* metric )
{
	delete metric;
}

extern "C" __declspec(dllexport) int GetVQMTVersion()
{
	return 320;
}

extern "C" __declspec(dllexport) int CompatibleWithVQMT(int vqmtVer)
{
	return vqmtVer >= 320 ? 0 : -1;
}

/*
* Initialization
*/

void CBIPSNRPlugin::Init ( IMetricPlugin::COLORCOMPONENT cc, int width, int height, int &global_id )
{
	m_bipsnr = global_id++;
	m_bimse = global_id++;

	m_width = width;
	m_height = height;
	m_size = width * height;

	m_frames_num = 0;

	m_mse_summ = 0.0f;
}

/*
* Measurement
*/

vector< pair <IMetricPlugin::ID, float> > CBIPSNRPlugin::Measure ( vector<IMetricImage*> &images )
{
	float * plane0 = images [ 0 ]->GetY ();
	float * plane1 = images [ 1 ]->GetY ();

	/*
	* Building brightness table
	*/	
	memset ( m_table, 0, sizeof ( m_table ) );

	for (int i = 0; i < m_size; i++)
    {
        int plane0_ix = (int)(plane0[i] + 0.5);
        if (plane0_ix < 0) plane0_ix = 0;
        if (plane0_ix > 255) plane0_ix = 255;
        int plane1_ix = (int)(plane1[i] + 0.5);
        if (plane1_ix < 0) plane1_ix = 0;
        if (plane1_ix > 255) plane1_ix = 255;
        m_table[plane0_ix][plane1_ix]++;
    }

	/*
	* Calculating m_table_transform values. They maximize average PSNR per frame.
	*/
	float psnr, mse = 0.0f;

	for ( int i = 0; i < 256; ++ i )
	{
		/*
		* Calculating helper values
		*/
		float summ_d = 0.0f; // Doubled sum of values in row
		float summ_w = 0.0f; // Weighted sum of values in row with weight values ( 2 * j + 1 )

		for ( int j = 0; j < 256; ++ j )
		{
			summ_d += m_table [ i ][ j ];
			summ_w += m_table [ i ][ j ] * j;
		}

		summ_w = 2 * summ_w + summ_d;
		summ_d = 2 * summ_d;

		/*
		* Calculating cost for value 0. Cost is sum of square differences of values which corresponds to value i
		*/
		float cost = 0.0f;

		for ( int j = 0; j < 256; ++ j ) cost += m_table [ i ][ j ] * float( sqr ( j ) );

		/*
		* Searching for optimal value
		*/
		int opt_val = 0;
		float opt_cost = cost;

		for ( int val = 1; val < 256; ++ val )
		{
			/*
			* Calculate next cost by iterative formula
			*/
			cost += val * summ_d - summ_w;

			if ( cost < opt_cost ) {
				opt_val = val;
				opt_cost = cost;
			}
		}

		m_table_transform [ i ] = opt_val;
		mse += opt_cost;
	}

	mse /= m_size;
	psnr = MSEToPSNR ( mse );

	m_mse_summ += mse;
	m_frames_num ++;

	vector< pair <IMetricPlugin::ID, float> > res;

	res.push_back ( make_pair ( m_bipsnr, psnr ) );
	res.push_back ( make_pair ( m_bimse, mse ) );

	return res;
}

vector< pair <IMetricPlugin::ID, float> > CBIPSNRPlugin::MeasureAndVisualize ( vector<IMetricImage*>&images, BYTE *vis, int vis_pitch )
{
	vector< pair <IMetricPlugin::ID, float> > res = Measure ( images );

	/*
	* Visualizing
	*/	
	
	int i, j;
	BYTE * cur_vis = vis;
	float * cur_plane0 = images [ 0 ]->GetY ();
	float * cur_plane1 = images [ 1 ]->GetY ();

	for ( i = 0; i < m_height; ++ i ) 
    {
        for ( j = 0; j < m_width; ++ j ) 
	    {
		    float diff = sqr ( *cur_plane1 - m_table_transform [ int ( *cur_plane0 + 0.5 ) ] );

		    if ( vis ) 
		        VisualizePSNR ( MSEToPSNR ( diff ), cur_vis + 3*j, cur_vis + 3*j + 1, cur_vis + 3*j + 2 );

		    cur_plane0 ++;
		    cur_plane1 ++;
	    }
        cur_vis += vis_pitch;
    }

	if ( vis ) 
	{
		for ( int i = 0; i < 256; ++ i )
		{		
			SetRGB ( vis, vis_pitch, i, m_height - i - 1, 200, 0, 0 ); // Red line
            // vertical grid
			SetRGB ( vis, vis_pitch, 63, m_height - i - 1, 128, 128, 128 );
			SetRGB ( vis, vis_pitch, 127, m_height - i - 1, 128, 128, 128 );
			SetRGB ( vis, vis_pitch, 191, m_height - i - 1, 128, 128, 128 );
			SetRGB ( vis, vis_pitch, 255, m_height - i - 1, 200, 200, 200 );
            // horizontal grid
			SetRGB ( vis, vis_pitch, i, m_height - 64, 128, 128, 128 );
			SetRGB ( vis, vis_pitch, i, m_height - 128, 128, 128, 128 );
			SetRGB ( vis, vis_pitch, i, m_height - 192, 128, 128, 128 );
			SetRGB ( vis, vis_pitch, i, m_height - 256, 200, 200, 200 );
		}
        // graph
		for ( int i = 0; i < 256; ++ i ) 
		{			
			for ( int j = 0; j < m_table_transform [ i ]; ++ j ) 
			    SetRGBA ( vis, vis_pitch, i, m_height - j - 1, 128, 128, 128, 0.4f );
			    
			for ( int j = m_table_transform [ i ] + 1; j < 256; ++ j ) 
			    SetRGBA ( vis, vis_pitch, i, m_height - j - 1, 0, 0, 0, 0.6f );

			SetRGB ( vis, vis_pitch, i, m_height - m_table_transform [ i ] - 1, 127, 255, 127 );
		}
	}

	return res;
}

vector< pair <IMetricPlugin::ID, float> > CBIPSNRPlugin::CalculateAverage ( bool visualize )
{
	vector< pair <IMetricPlugin::ID, float> > res;


	float mse = m_mse_summ / m_frames_num;

	res.push_back ( make_pair ( m_bipsnr, MSEToPSNR ( mse ) ) );
	res.push_back ( make_pair ( m_bimse, mse ) );

	return res;
}

/*
* Private methods
*/

void CBIPSNRPlugin::SetRGB ( BYTE * img, int pitch, int x, int y, BYTE r, BYTE g, BYTE b )
{
    if (x < 0 || m_width <= x || y < 0 || m_height <= y)
    {
        return;
    }
	int ofs = y * pitch + x * 3;

	img [ ofs + 2 ] = r;
	img [ ofs + 1 ] = g;
	img [ ofs + 0 ] = b;
}

void CBIPSNRPlugin::SetRGBA ( BYTE * img, int pitch, int x, int y, float r, float g, float b, float a )
{
    if (x < 0 || m_width <= x || y < 0 || m_height <= y)
    {
        return;
    }
	int ofs = y * pitch + x * 3;

	img [ ofs + 2 ] = BYTE ( r * a + img [ ofs + 2 ] * ( 1 - a ) );
	img [ ofs + 1 ] = BYTE ( g * a + img [ ofs + 1 ] * ( 1 - a ) );
	img [ ofs + 0 ] = BYTE ( b * a + img [ ofs + 0 ] * ( 1 - a ) );
}


void CBIPSNRPlugin::VisualizePSNR ( float metr_val, BYTE *b, BYTE *g, BYTE *r )
{
	/*	
	* Visualization map is the same as in internal PSNR metric 
	*/
	*b = *g = *r = 0;
	if(metr_val >= 99)
		return;
	else if(metr_val >= 35)
		*b = max(255 - (short)(metr_val - 25) * 6, 40);
	else if(metr_val >= 20 && metr_val < 35)
		*g = 255 - (short)(metr_val - 20) * 15;								
	else if(metr_val >= 17 && metr_val < 20)
		*g = *r = (255 - short(metr_val - 17) * 5);
	if(metr_val >= 0 && metr_val < 17)
		*r = 255 - (short)(metr_val) * 3;
	return;
}

/*
* Calculates value of PSNR corresponding to MSE
*/
float CBIPSNRPlugin::MSEToPSNR ( float p_mse )
{
	if ( p_mse != 0.0f ) {
		return min ( 10.0f * log10 ( 65025.0f / p_mse ), 100.0f );
	} else {
		return 100.0f;
	}
}

/*
* Config
*/

vector< pair <IMetricPlugin::ID, pair< vector <int>, wstring> > > CBIPSNRPlugin::MapIDToFrame ( bool visualize )
{
	vector< pair <IMetricPlugin::ID, pair<vector <int>, wstring> > > res;

	/*
	* frames - vector of numbers of frames that given ID corresponds to. 
	* Here we have two IDs, both of which correspond to both frames 
	*/
	vector<int> frames;
	frames.push_back(0);
	frames.push_back(1);

	res.push_back ( make_pair ( m_bipsnr, make_pair ( frames, L"MSU BI-PSNR" ) ) );
	res.push_back ( make_pair ( m_bimse,  make_pair ( frames, L"MSU BI-MSE"  ) ) );
	return res;
}

int CBIPSNRPlugin::GetVideoNum ( bool visualize )
{
	return 2;
}

vector <IMetricPlugin::COLORCOMPONENT> CBIPSNRPlugin::GetSupportedColorcomponents ()
{
	vector <IMetricPlugin::COLORCOMPONENT> cc;
	cc.push_back(IMetricPlugin::YYUV);
	return cc;
}

wstring CBIPSNRPlugin::GetName ()
{
	return L"bi_psnr";
}

wstring CBIPSNRPlugin::GetUnit ()
{
	return L"dB";
}

int CBIPSNRPlugin::GetWidthMultiply ()
{
	return 1;
}

int CBIPSNRPlugin::GetHeightMultiply ()
{
	return 1;
}

wstring	CBIPSNRPlugin::GetMetrInfoURL ()
{
	return L"http://www.compression.ru/video/quality_measure/metric_plugins/bi-psnr_en.htm";
}

bool CBIPSNRPlugin::GetMetrIncline ()
{
	return true;
}

wstring	CBIPSNRPlugin::GetInterfaceName ()
{
	return L"MSU Brightness Independent PSNR v1.1";
}

bool CBIPSNRPlugin::IsConfigurable ()
{
	return false;
}

bool CBIPSNRPlugin::Config ( const wstring &configuration_string )
{
	return true;
}

void CBIPSNRPlugin::GetDefaultConfiguration ( wstring &def_config )
{
	return;
}

//EOF//////////////////////////////////////////////////////////////////////////
